import torch
import torch.utils.data


class ClassificationAugmentationMixin:
    def augment(self, image):
        if self.augment_fn is not None:
            # image = self.augment_fn(image=image)
            # # TODO: curry albumentations function so we don't have to do this
            # try:
            #     image = image["image"]
            # except KeyError:
            #     pass
            image = self.augment_fn(image=image)["image"]

        return image


class ClassificationTTAMixin:
    def tta(self, image, label=None):
        if self.tta_fn is not None:
            image, label = self.tta_fn(image, label)
        else:
            pass
        return image, label


class ImageClassificationDatasetV2(
    torch.utils.data.Dataset, ClassificationAugmentationMixin, ClassificationTTAMixin
):
    def __init__(
        self, samples, augment_fn: lambda x: x, convert_image_to_tensor_fn=lambda x: x
    ):
        self.samples = samples
        self.augment_fn = augment_fn
        self.convert_image_to_tensor_fn = convert_image_to_tensor_fn

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, i):
        sample = self.samples[i]
        image, label = sample.data

        image = self.augment(image)
        tensor = self.convert_image_to_tensor_fn(image)

        if sample.has_label:
            sample.label.tensor = label
        sample.image.tensor = tensor
        return sample


class ImageClassificationDataset(
    torch.utils.data.Dataset, ClassificationAugmentationMixin, ClassificationTTAMixin
):
    def __init__(
        self,
        samples_data,
        create_sample_fn=lambda x: x,
        augment_fn=lambda x: x,
        convert_image_to_tensor_fn=lambda x: x,
    ):
        self.samples_data = samples_data
        self.create_sample_fn = create_sample_fn
        self.augment_fn = augment_fn
        self.convert_image_to_tensor_fn = convert_image_to_tensor_fn

    def __len__(self):
        return len(self.samples_data)

    def __getitem__(self, i):
        sample = self.create_sample_fn(self.samples_data[i])
        image, label = sample.data

        image = self.augment(image)
        tensor = self.convert_image_to_tensor_fn(image)

        if sample.has_label:
            sample.label.tensor = label
        sample.image.tensor = tensor
        return sample


def image_classification_collate_fn(samples):
    batch = {}

    batch["names"] = torch.utils.data.dataloader.default_collate(
        [s.name for s in samples]
    )

    batch["images"] = torch.stack(
        [s.image.tensor.type(torch.FloatTensor) for s in samples]
    )

    labels_exists = samples[0].has_label

    if labels_exists:
        batch["labels"] = torch.utils.data.dataloader.default_collate(
            [s.label.tensor for s in samples]
        )
    return batch


lookup = {
    "ImageClassificationDataset": ImageClassificationDataset,
    "image_classification_collate_fn": image_classification_collate_fn,
    "ImageClassificationDatasetV2": ImageClassificationDatasetV2,
}



from .config_loader import (
    dump_config_to_path,
    dump_config_to_string,
    load_config_from_path,
    load_config_from_string,
    load_config_from_strings,
    merge_dicts,
)
from .config_recorder import ConfigurationNode, ConfigurationNodeRecorder, create_output_names, patch_output_value
from .graph_construction import find_references, flatten_resources_dict, replace_arguments, replace_references, replace_should_run
from .node import Node, NodeRemoveMode, remove_node_arguments
from .pipeline import Pipeline
from .utils import *
from .yaml_loader import Reference, Variable



from pathlib import Path

from torch.utils.tensorboard import SummaryWriter

from ....logger import logger
from ....utils.convenience import catch_all_exceptions
from .. import Callback


class TensorboardRecorder(Callback):
    _order = -10

    def __init__(self, root_save_path, name=None):
        super().__init__(name)
        self.tensorboard_save_path = Path(root_save_path) / "tensorboard"
        self._writer = None

    def getWriter(self):
        if self._writer is None:
            self._writer = SummaryWriter(log_dir=self.tensorboard_save_path)
            logger.info(f"Run tensorboard --logdir {str(self.tensorboard_save_path)}")
        return self._writer

    def on_train_batch_end(self, context):
        self._record_loss(context.loss, context.current_phase, context.current_step)

    def on_validation_epoch_begin(self, context):
        self.validation_step_offset = 0

    def on_validation_batch_end(self, context):
        self._record_loss(
            context.loss,
            context.current_phase,
            context.current_step + self.validation_step_offset,
        )
        self.validation_step_offset += 1

    def on_validation_epoch_end(self, context):
        self._record_metric_on_epoch_end(
            context.validation_metric,
            context.current_phase,
            context.current_step + self.validation_step_offset,
        )

    def on_test_batch_end(self, context):
        self._record_loss(context.loss, context.current_phase, context.current_step)

    def on_test_epoch_end(self, context):
        self._record_metric_on_epoch_end(
            context.test_metric, context.current_phase, context.current_step
        )

    def _record_loss(self, losses, phase, step):
        phase = str(phase)
        all_losses = {"Total": losses.total_loss.item()}
        for loss_name, loss_value in losses.reduced_losses.items():
            key = f"{phase}/{loss_name}"
            loss = loss_value.item()
            self.getWriter().add_scalar(key, loss, step)
            all_losses[key] = loss
        self.getWriter().add_scalars(f"{str(phase)}/All Losses", all_losses, step)

    def _record_metric_on_epoch_end(self, metrics, phase, step):
        if metrics is None:
            return

        for metric in metrics.metrics:
            metric.write_to_tensorboard_on_epoch_end(self.getWriter(), phase, step)



from collections import defaultdict

from ....utils.convenience import to_numpy
from ..callback import Callback


class ImageClassificationResultRecorder(Callback):

    _order = -10

    def __init__(self, convert_tensor_to_image_fn=None, name=None):
        super().__init__(name)
        self.convert_tensor_to_image_fn = convert_tensor_to_image_fn
        self.results = defaultdict(list)

    def on_train_batch_end(self, context):
        phase, epoch, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.train_batch,
            context.train_prediction,
        )
        images = self.convert_tensor_to_image_fn(batch["images"])
        predictions = to_numpy(predictions)
        labels = to_numpy(batch["labels"])
        self.results[str(phase), epoch].extend(
            [
                {"image": image, "prediction": prediction, "label": label}
                for image, prediction, label in zip(images, predictions, labels)
            ]
        )

    def on_validation_batch_end(self, context):
        phase, epoch, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.validation_batch,
            context.validation_prediction,
        )
        images = self.convert_tensor_to_image_fn(batch["images"])
        predictions = to_numpy(predictions)
        labels = to_numpy(batch["labels"])
        self.results[str(phase), epoch].extend(
            [
                {"image": image, "prediction": prediction, "label": label}
                for image, prediction, label in zip(images, predictions, labels)
            ]
        )

    def on_test_batch_end(self, context):
        phase, epoch, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.test_batch,
            context.test_prediction,
        )
        images = self.convert_tensor_to_image_fn(batch["images"])
        predictions = to_numpy(predictions)
        labels_exists = "labels" in batch

        if labels_exists:
            labels = to_numpy(batch["labels"])
            results = [
                {"image": image, "prediction": prediction, "label": label}
                for image, prediction, label in zip(images, predictions, labels)
            ]
        else:
            results = [
                {"image": image, "prediction": prediction}
                for image, prediction in zip(images, predictions)
            ]

        self.results[str(phase), epoch].extend(results)


class ImageClassificationInferenceRecorder(Callback):

    _order = -10

    def __init__(self, convert_tensor_to_image_fn=None, name=None):
        super().__init__(name)
        self.convert_tensor_to_image_fn = convert_tensor_to_image_fn
        self.results = []

    @property
    def serializable_results(self):
        serializable_results = []
        for result in self.results:
            serializable_results.append({k: v.tolist() for k, v in result.items()})
        return serializable_results

    def flush_results(self):
        self.results = []

    def on_test_batch_end(self, context):
        _, _, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.test_batch,
            context.test_prediction,
        )
        images = self.convert_tensor_to_image_fn(batch["images"])
        predictions = to_numpy(predictions)
        self.results.extend(
            [
                {"image": image, "prediction": prediction}
                for image, prediction in zip(images, predictions)
            ]
        )



from copy import deepcopy

import numpy as np


def all_methods_in_callback_handler_are_called(callback_handler):
    return all(
        callback_handler.get_callback_by_name(
            "CallbackMethodIsCalled"
        ).method_is_called.values()
    )


def arrays_are_same(array1, array2):
    return np.all(array1 == array2)


def copy_and_flatten_model_weights(model):
    return deepcopy([p.cpu().detach().numpy().flatten() for p in model.parameters()])


lookup = {
    "all_methods_in_callback_handler_are_called": all_methods_in_callback_handler_are_called,
    "arrays_are_same": arrays_are_same,
    "copy_and_flatten_model_weights": copy_and_flatten_model_weights,
}



import pytorch_lightning as pl

from ......core.callbacks import CallbackHandlerContextValues
from ......utils.convenience import debug_log_on_call, timing, to_numpy


class PytorchLightningImageClassificationModule(pl.LightningModule):
    @debug_log_on_call
    def __init__(
        self,
        model_creator,
        optimizer_creator,
        train_samples,
        train_dataset_creator,
        train_dataloader_creator,
        loss,
        val_samples=None,
        val_dataset_creator=None,
        val_dataloader_creator=None,
        test_samples=None,
        test_dataset_creator=None,
        test_dataloader_creator=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.model = model_creator()
        self.optimizer_creator = optimizer_creator
        self.loss = loss

        self.train_samples = train_samples
        self.train_dataset_creator = train_dataset_creator
        self.train_dataloader_creator = train_dataloader_creator

        self.val_samples = val_samples
        self.val_dataset_creator = val_dataset_creator
        self.val_dataloader_creator = val_dataloader_creator

        self.test_samples = test_samples
        self.test_dataset_creator = test_dataset_creator
        self.test_dataloader_creator = test_dataloader_creator

    @debug_log_on_call
    def _can_train(self):
        return (
            self.train_samples is not None
            and self.train_dataset_creator is not None
            and self.train_dataloader_creator is not None
        )

    @debug_log_on_call
    def _can_val(self):
        return (
            self.val_samples is not None
            and self.val_dataset_creator is not None
            and self.val_dataloader_creator is not None
        )

    @debug_log_on_call
    def _can_test(self):
        return (
            self.test_samples is not None
            and self.test_dataset_creator is not None
            and self.test_dataloader_creator is not None
        )

    @debug_log_on_call
    def prepare_data(self):
        if self._can_train():
            self.train_dataset = self.train_dataset_creator(samples=self.train_samples)
        if self._can_val():
            self.val_dataset = self.val_dataset_creator(samples=self.val_samples)
        if self._can_test():
            self.test_dataset = self.test_dataset_creator(samples=self.test_samples)

    @debug_log_on_call
    def train_dataloader(self):
        return self.train_dataloader_creator(self.train_dataset)

    @debug_log_on_call
    def val_dataloader(self):
        if self._can_val():
            return self.val_dataloader_creator(self.val_dataset)
        else:
            return None

    @debug_log_on_call
    def test_dataloader(self):
        if self._can_test():
            return self.test_dataloader_creator(self.test_dataset)
        else:
            return None

    @debug_log_on_call
    def configure_optimizers(self):
        return self.optimizer_creator(self.model.parameters())

    @debug_log_on_call
    def training_step(self, batch, batch_idx):
        x = batch["images"]
        y = batch["labels"]
        prediction = self(x)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TRAIN_PREDICTION, prediction
        )
        prediction = self.callback_handler.on_train_loss_begin()
        loss = self.loss(prediction, y)
        return {"loss": loss}

    @debug_log_on_call
    def forward(self, x):
        return self.model(x)

    @debug_log_on_call
    def validation_step(self, batch, batch_idx):
        x = batch["images"]
        y = batch["labels"]
        prediction = self.model(x)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.VALIDATION_PREDICTION, prediction
        )
        prediction = self.callback_handler.on_validation_loss_begin()
        loss = self.loss(prediction, y)
        return {"val_loss": loss}

    @debug_log_on_call
    def test_step(self, batch, batch_idx):
        x = batch["images"]
        prediction = self.model(x)

        output = {"prediction": to_numpy(prediction)}
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TEST_PREDICTION, prediction
        )
        prediction = self.callback_handler.on_test_loss_begin()

        if "labels" in batch:
            y = batch["labels"]
            loss = self.loss(prediction, y)
            output.update({"test_loss": loss})
        return output


def create_pytorch_lightning_image_classification_module(
    callback_handler,
    model_creator,
    optimizer_creator=None,
    loss=None,
    train_samples=None,
    train_dataset_creator=None,
    train_dataloader_creator=None,
    val_samples=None,
    val_dataset_creator=None,
    val_dataloader_creator=None,
    test_samples=None,
    test_dataset_creator=None,
    test_dataloader_creator=None,
):
    from ......integrations.pytorch_lightning.adapters import PytorchLightningModuleHooksToCallbackHandlerAdapter

    return PytorchLightningModuleHooksToCallbackHandlerAdapter(
        PytorchLightningImageClassificationModule(
            model_creator=model_creator,
            optimizer_creator=optimizer_creator,
            loss=loss,
            train_samples=train_samples,
            train_dataset_creator=train_dataset_creator,
            train_dataloader_creator=train_dataloader_creator,
            val_samples=val_samples,
            val_dataset_creator=val_dataset_creator,
            val_dataloader_creator=val_dataloader_creator,
            test_samples=test_samples,
            test_dataset_creator=test_dataset_creator,
            test_dataloader_creator=test_dataloader_creator,
        ),
        callback_handler,
    )


lookup = {
    "PytorchLightningImageClassificationModule": PytorchLightningImageClassificationModule,
    "create_pytorch_lightning_image_classification_module": create_pytorch_lightning_image_classification_module,
}



from ....utils.convenience import to_numpy
from .. import Callback
from .utils import Storage


class PerSampleLossRecorder(Callback):
    _order = -10

    def __init__(self, name=None):
        super().__init__(name)
        self.losses = Storage()

    def _store_losses(self, batch, loss, phase, epoch, step):
        for loss_name, loss_values in loss.per_sample_losses.items():
            detached_loss_values = to_numpy(loss_values).tolist()
            if not isinstance(detached_loss_values, list):
                detached_loss_values = [detached_loss_values]
            self.losses.write(str(phase), epoch, loss_name, detached_loss_values)

    def on_train_batch_end(self, context):
        self._store_losses(
            context.train_batch,
            context.loss,
            context.current_phase,
            context.current_epoch,
            context.current_step,
        )

    def on_validation_batch_end(self, context):
        self._store_losses(
            context.validation_batch,
            context.loss,
            context.current_phase,
            context.current_epoch,
            context.current_step,
        )

    def on_test_batch_end(self, context):
        self._store_losses(
            context.test_batch,
            context.loss,
            context.current_phase,
            context.current_epoch,
            context.current_step,
        )



import numpy as np
from p_tqdm import p_map
from tqdm import tqdm

from ..data import create_image_bboxes_detection_sample


def load_label_from_path_ultralytics_format(path):
    labels = np.loadtxt(path).reshape(-1, 5)

    if len(labels) == 0:
        return np.array([])

    return np.array(labels[:, 0].astype(np.int32))


def load_bboxes_from_path_ultralytics_format(path):
    bboxes = np.loadtxt(path).reshape(-1, 5)

    if len(bboxes) == 0:
        return np.array([])

    return np.array(bboxes[:, 1:])


def create_bboxes_detection_samples_data_from_ultralytics_format(
    image_paths,
    labels_paths,
    bboxes_format,
    transform_to_bboxes_format=None,
    labels_to_keep=None,
    transform_label_map=None,
):
    return np.array(
        [
            create_bboxes_detection_sample_data_from_ultralytics_format(
                image_path,
                labels_path,
                bboxes_format,
                transform_to_bboxes_format,
                labels_to_keep,
                transform_label_map,
            )
            for image_path, labels_path in zip(image_paths, labels_paths)
        ]
    )


def create_bboxes_detection_sample_data_from_ultralytics_format(
    image_path,
    labels_path,
    bboxes_format,
    transform_to_bboxes_format=None,
    labels_to_keep=None,
    transform_label_map=None,
):
    return {
        "image_path": image_path,
        "labels_path": labels_path,
        "load_labels_fn": load_label_from_path_ultralytics_format,
        "bboxes_path": labels_path,
        "load_bboxes_fn": load_bboxes_from_path_ultralytics_format,
        "bboxes_format": bboxes_format,
        "transform_to_bboxes_format": transform_to_bboxes_format,
        "labels_to_keep": labels_to_keep,
        "transform_label_map": transform_label_map,
    }


def create_bboxes_detection_samples_from_ultralytics_format(
    image_paths,
    labels_paths,
    bboxes_format,
    transform_to_bboxes_format,
    labels_to_keep=None,
    transform_label_map=None,
    first_n_paths=None,
    parallel=True,
):
    assert len(image_paths) == len(labels_paths)
    if first_n_paths is None:
        first_n_paths = len(image_paths)

    if parallel:
        samples = p_map(
            _create_image_bboxes_detection_sample_lambda,
            create_bboxes_detection_samples_data_from_ultralytics_format(
                image_paths[:first_n_paths],
                labels_paths[:first_n_paths],
                bboxes_format,
                transform_to_bboxes_format,
                labels_to_keep,
                transform_label_map,
            ),
        )
    else:
        samples = []
        for image_path, labels_path in tqdm(zip(image_paths, labels_paths)):
            samples.append(
                _create_image_bboxes_detection_sample_lambda(
                    create_bboxes_detection_sample_data_from_ultralytics_format(
                        image_path,
                        labels_path,
                        bboxes_format,
                        transform_to_bboxes_format,
                        labels_to_keep,
                        transform_label_map,
                    )
                ),
            )

    return np.array(samples)


def _create_image_bboxes_detection_sample_lambda(sample_data):
    return create_image_bboxes_detection_sample(**sample_data)


lookup = {
    "load_bbox_from_path_ultralytics_format": load_bboxes_from_path_ultralytics_format,
    "load_label_from_path_ultralytics_format": load_label_from_path_ultralytics_format,
    "create_bboxes_detection_samples_from_ultralytics_format": create_bboxes_detection_samples_from_ultralytics_format,
    "create_bboxes_detection_samples_data_from_ultralytics_format": create_bboxes_detection_samples_data_from_ultralytics_format,
}



import pytorch_lightning as pl

from ....core.callbacks import CallbackHandler, CallbackHandlerContextValues
from ....utils.convenience import debug_log_on_call


class PytorchLightningCallbacksToCallbackHandlerAdapter(pl.Callback):
    def __init__(self, callback_handler: CallbackHandler):
        self.callback_handler = callback_handler

    @debug_log_on_call
    def setup(self, trainer, model, stage):
        """Called in the beginning of fit and test"""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TRAINER, trainer
        )
        self.callback_handler.on_setup()

    @debug_log_on_call
    def teardown(self, trainer, model, stage):
        """Called at the end of fit and test"""
        self.callback_handler.on_teardown()

    @debug_log_on_call
    def on_init_start(self, trainer):
        """Called when the trainer initialization begins, model has not yet been set."""
        self.callback_handler.on_init_begin()

    @debug_log_on_call
    def on_init_end(self, trainer):
        """Called when the trainer initialization ends, model has not yet been set."""
        self.callback_handler.on_init_end()

    @debug_log_on_call
    def on_fit_start(self, trainer, model):
        """Called when the trainer initialization begins, model has not yet been set."""
        pass

    @debug_log_on_call
    def on_fit_end(self, trainer, model):
        """Called when the trainer initialization begins, model has not yet been set."""
        pass

    def on_sanity_check_start(self, trainer, model):
        """Called when the validation sanity check starts."""
        pass

    def on_sanity_check_end(self, trainer, model):
        """Called when the validation sanity check ends."""
        pass

    def on_train_epoch_start(self, trainer, model):
        """Called when the epoch begins."""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.CURRENT_EPOCH, trainer.current_epoch
        )
        self.callback_handler.on_train_epoch_begin()

    def on_train_epoch_end(self, trainer, model):
        """Called when the epoch ends."""
        pass

    def on_validation_epoch_start(self, trainer, model):
        """Called when the epoch begins."""
        self.callback_handler.on_validation_epoch_begin()

    def on_validation_epoch_end(self, trainer, model):
        """Called when the epoch ends."""
        pass

    def on_test_epoch_start(self, trainer, model):
        """Called when the epoch begins."""
        self.callback_handler.on_test_epoch_begin()
        pass

    def on_test_epoch_end(self, trainer, model):
        """Called when the epoch ends."""
        pass

    def on_epoch_start(self, trainer, model):
        """Called when the epoch begins."""
        pass

    def on_epoch_end(self, trainer, model):
        """Called when the epoch ends."""
        pass

    def on_train_start(self, trainer, model):
        """Called when the train begins."""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.MODEL, model
        )
        self.callback_handler.on_train_begin()

    def on_train_end(self, trainer, model):
        """Called when the train ends."""
        self.callback_handler.on_train_end()
        pass

    def on_pretrain_routine_start(self, trainer, model):
        """Called when the train begins."""
        pass

    def on_pretrain_routine_end(self, trainer, model):
        """Called when the train ends."""
        pass

    def on_batch_start(self, trainer, model):
        """Called when the training batch begins."""
        pass

    def on_batch_end(self, trainer, model):
        """Called when the training batch ends."""
        pass

    def on_train_batch_start(self, trainer, model, batch, batch_idx, dataloader_idx):
        """Called when the training batch begins."""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.CURRENT_STEP, trainer.global_step
        )
        pass

    def on_train_batch_end(
        self, trainer, model, outputs, batch, batch_idx, dataloader_idx
    ):
        """Called when the training batch ends."""
        self.callback_handler.on_train_batch_end()

    def on_validation_batch_start(
        self, trainer, model, batch, batch_idx, dataloader_idx
    ):
        """Called when the validation batch begins."""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.CURRENT_STEP, trainer.global_step
        )
        pass

    def on_validation_batch_end(
        self, trainer, model, outputs, batch, batch_idx, dataloader_idx
    ):
        """Called when the validation batch ends."""
        self.callback_handler.on_validation_batch_end()

    def on_test_batch_start(self, trainer, model, batch, batch_idx, dataloader_idx):
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.CURRENT_STEP, trainer.global_step
        )
        """Called when the test batch begins."""
        pass

    def on_test_batch_end(
        self, trainer, model, outputs, batch, batch_idx, dataloader_idx
    ):
        """Called when the test batch ends."""
        self.callback_handler.on_test_batch_end()

    def on_validation_start(self, trainer, model):
        """Called when the validation loop begins."""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.MODEL, model
        )
        self.callback_handler.on_validation_begin()

    def on_validation_end(self, trainer, model):
        """Called when the validation loop ends."""
        self.callback_handler.on_validation_end()

    def on_test_start(self, trainer, model):
        """Called when the test begins."""
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.MODEL, model
        )
        self.callback_handler.on_test_begin()

    def on_test_end(self, trainer, model):
        """Called when the test ends."""
        self.callback_handler.on_test_end()

    def on_keyboard_interrupt(self, trainer, model):
        """Called when the training is interrupted by KeyboardInterrupt."""
        pass


lookup = {
    "PytorchLightningCallbacksToCallbackHandlerAdapter": PytorchLightningCallbacksToCallbackHandlerAdapter
}



import torch.nn as nn


def conv2d_relu(ni, nf, ks=3, stride=1, padding=None, bn=False, bias=False):
    """Create a `conv2d` layer with `nn.ReLU` activation and optional(`bn`) `nn.BatchNorm2d`: `ni` input, `nf` out
    filters, `ks` kernel, `stride`:stride, `padding`:padding, `bn`: batch normalization."""
    layers = [
        conv2d(ni, nf, ks=ks, stride=stride, padding=padding, bias=bias),
        nn.ReLU(),
    ]
    if bn:
        layers.append(nn.BatchNorm2d(nf))
    return nn.Sequential(*layers)


def conv2d_trans(ni, nf, ks=2, stride=2, padding=0):
    "Create `nn.ConvTranspose2d` layer: `ni` inputs, `nf` outputs, `ks` kernel size, `stride`: stride. `padding` defaults to 0."
    return nn.ConvTranspose2d(ni, nf, kernel_size=ks, stride=stride, padding=padding)


def conv2d(ni, nf, ks=3, stride=1, padding=None, bias=False):
    "Create `nn.Conv2d` layer: `ni` inputs, `nf` outputs, `ks` kernel size. `padding` defaults to `k//2`."
    if padding is None:
        padding = ks // 2
    return nn.Conv2d(ni, nf, kernel_size=ks, stride=stride, padding=padding, bias=bias)
